#include "autoptr"
#include <cmath>

typedef enum dyntype
{
    DYNTYPE_VALUE,
    DYNTYPE_PLUSMINUS,
    DYNTYPE_TIMESDIV,
    DYNTYPE_MOD,
    DYNTYPE_FUNC
} dyntype;

class dynerror
{
public:
    enum type { getunsetval=1, getunsetexp, virtualcall };
private:
    type typi;
public:
    explicit dynerror(type t) : typi(t) { }
    operator const char * () const
    {
        return typi==getunsetval ? "Tried to get a value of UNSET variable"
             : typi==getunsetexp ? "Tried to get expression of UNSET variable"
             : typi==virtualcall ? "Pure virtual function called"
             : "Failing dyn error";
    }
};

#include <iostream>
std::ostream &operator<< (std::ostream &dest, dyntype t)
{
    switch(t)
    {
        case DYNTYPE_VALUE: dest << "VALUE"; break;
        case DYNTYPE_PLUSMINUS: dest << "PLUSMINUS"; break;
        case DYNTYPE_TIMESDIV: dest << "TIMESDIV"; break;
        case DYNTYPE_MOD: dest << "MOD"; break;
        case DYNTYPE_FUNC: dest << "FUNC"; break;
        default: dest << '?' << (int)t << '?';
    }
    return dest;
}

/* This is the base type of the expressions */
template<typename T>
class dyni : public ptrable
{
public:
    virtual ~dyni()
    {
    	std::cout << "dyni dying: \"" << std::flush;
    	/*debug();*/ std::cout << '"' << std::endl;
    }
    virtual const T getvalue() const = 0;
    virtual void debug() const = 0; //{ throw dynerror(dynerror::virtualcall); }
    virtual const dyntype gettype() const = 0;
};

/* Expression type: Constant */
template<typename T>
class dynvalue : public dyni<T>
{
    T v;
public:
    explicit dynvalue(const T &c) : v(c) { }
    virtual const T getvalue() const
    {
        return v;
    }
    virtual void debug() const { std::cout << v << std::flush; }
    virtual const dyntype gettype() const { return DYNTYPE_VALUE; }
};

/* This is the public type */
template<typename T>
class dyn : public ptrable
{
protected:
    autoptr<dyni<T> > v;
    long pp;
public:
    explicit dyn(autoptr<dyni<T> > p = NULL) : v(p), pp(0) { }
    dyn(const T &c) : v(new dynvalue<T> (c)), pp(0) { }
    //dyn(const dyn &p) : ptrable(), v(p.v), pp(p.pp) { }
    virtual ~dyn()
    {
    	std::cout << "dyn dying: \"" << std::flush;
    	debug(); std::cout << '"' << std::endl;
    }
    
    const T getvalue() const
    {
        if(!v) throw dynerror(dynerror::getunsetval);
        return v->getvalue() + pp;
    }
    void debug() const
    {
        if(!v) throw dynerror(dynerror::getunsetval);
        if(pp != 0)std::cout << '[';
        v->debug();
        if(pp != 0)
        {
            if(pp > 0)std::cout << '+';
            std::cout << pp;
        }
        if(pp != 0)std::cout << ']';
    }
    dyn &operator= (const dyn<T> &p)            { v = p.v; return *this; }
    dyn &operator= (const autoptr<dyni<T> > &p) { v = p;   return *this; }
    dyn &operator= (dyni<T> *p)                 { v = p;   return *this; }
    
    dyn &operator= (const T &value); // constant
    dyn operator++ () { ++pp; return *this; }
    dyn operator-- () { --pp; return *this; }
    dyn operator++ (int) { dyn tmp = *this; ++pp; return tmp; }
    dyn operator-- (int) { dyn tmp = *this; --pp; return tmp; }
    
    /* typecast operator */
    //operator T () const { return getvalue(); }
    
    const dyni<T> *const getexpr() const
    {
        if(!v) throw dynerror(dynerror::getunsetexp);
        return v;
    }
};

template<typename T>
dyn<T> &dyn<T>::operator= (const T &value)
{
    return *this = new dynvalue<T> (value);
}

template<typename T>
class dyn2pair : public ptrable
{
private:
    dyn2pair<T> &operator= (const dyn2pair<T> &) { }
public:
    bool nega;
    const dyn<T> &expr;
public:
    dyn2pair(bool neg, const dyn<T> &exp) : nega(neg), expr(exp)
    {
        std::cout << "Created: " << (neg?'-':'+') << ' ';
        exp.debug();
        std::cout << std::endl;
    }
    ~dyn2pair()
    {
        std::cout << "Killed: " << (nega?'-':'+') << ' ';
        expr.debug();
        std::cout << std::endl;
    }
};

#include <list>
template<typename T>
class dyn2coll : public dyni<T>
{
private:
    char sgneg, sgpos;
    typedef autoptr<dyn2pair<T> > elemtype;
    typedef std::list<elemtype> elemlist;
    typedef autoptr<dyn<T> > tmptype;
    std::list<tmptype> tmplist;
    void addtmp(dyn<T> *p)
    {
        /* Add p to the automatic temporary variable deletion list 
         * (They will be deleted when the destructor is called, due to autoptr.
         */
        tmplist.push_back(p);
    }
protected:
    elemlist elems;
public:
    dyn2coll(char n, char p) : sgneg(n), sgpos(p) { }
    virtual void debug() const
    {
        if(!elems.size()) { std::cout << "[EMPTY]"; return; }
        if(elems.size() > 1)std::cout << '(';
        elemlist::const_iterator i = elems.begin();
        (*i)->expr.debug();
        while(++i != elems.end())
        {
            std::cout << (char)((*i)->nega ? sgneg : sgpos);
            (*i)->expr.debug();
        }
        if(elems.size() > 1)std::cout << ')';
    }
    void add(dyn<T> *elem, bool nega)
    {
        addtmp(elem);
        add(*elem, nega);
    }
    void add(const dyn<T> &elem, bool nega);
};

template<typename T>
void dyn2coll<T>::add(const dyn<T> &elem, bool nega)
{
    const dyni<T> *expr = elem.getexpr();
    //std::cout << "ADD " << gettype() << ": "; debug(); std::cout << " <- " << expr->gettype() << ": "; expr->debug(); std::cout << std::endl;
    if(expr->gettype() == gettype())
    { 
        /* For unknown reason, dynamic_cast causes segmentation fault.
         * static_cast does not, and works as supposed.
         */
        const dyn2coll<T> *expc = static_cast<const dyn2coll<T> *> (expr);
        const elemlist e = expc->elems;
        elemlist::const_iterator i;
        for(i=e.begin(); i!=e.end(); ++i)
            elems.push_back(new dyn2pair<T> (nega ^ (*i)->nega, (*i)->expr));
    }
    else
        elems.push_back(new dyn2pair<T> (nega, elem));
    //std::cout << "DONE" << std::endl;
}


template<typename T>
class dynplusminus : public dyn2coll<T>
{
public:
    dynplusminus() : dyn2coll<T> ('-', '+') { }
    virtual const T getvalue() const;
    virtual const dyntype gettype() const { return DYNTYPE_PLUSMINUS; }
};

template<typename T>
class dyntimesdiv : public dyn2coll<T>
{
public:
    dyntimesdiv() : dyn2coll<T> ('/', '*') { }
    virtual const T getvalue() const;
    virtual const dyntype gettype() const { return DYNTYPE_TIMESDIV; }
};

template<typename T>
const T dynplusminus<T>::getvalue() const
{
    elemlist::const_iterator i = elems.begin();
    T result = (*i)->expr.getvalue();
    while(++i != elems.end())
        if((*i)->nega) result -= (*i)->expr.getvalue();
        else result += (*i)->expr.getvalue();
    return result;
}

template<typename T>
const T dyntimesdiv<T>::getvalue() const
{
    elemlist::const_iterator i = elems.begin();
    T result = (*i)->expr.getvalue();
    while(++i != elems.end())
        if((*i)->nega) result /= (*i)->expr.getvalue();
        else result *= (*i)->expr.getvalue();
    return result;
}

#define dyn2opdef2(op, opa, nega, obj) \
  template<typename T> \
  const dyn<T> operator op (const dyn<T> &l, const dyn<T> &r) \
  { \
    obj<T> *p = new obj<T>; \
    p->add(l, false); \
    p->add(r, nega); \
    return dyn<T> (p); \
  } \
  template<typename T> \
  const dyn<T> operator op (const dyn<T> &l, const T &value) \
  { \
    dyn<T> *tmp = new dyn<T> (new dynvalue<T> (value)); \
    obj<T> *p = new obj<T>; \
    p->add(l, false); \
    p->add(tmp, nega); \
    return dyn<T> (p); \
  } \
  template<typename T> \
  const dyn<T> &operator opa (dyn<T> &dest, const dyn<T> &value) \
  { \
    dyn<T> *tmp  = new dyn<T> (dest); \
    obj<T> *p = new obj<T>; \
    p->add(tmp, false); \
    p->add(value, nega); \
    dest = p; \
    return dest; \
  } \
  template<typename T> \
  const dyn<T> &operator opa (dyn<T> &dest, const T &value) \
  { \
    dyn<T> *tmp  = new dyn<T> (dest); \
    dyn<T> *tmp2 = new dyn<T> (new dynvalue<T> (value)); \
    obj<T> *p = new obj<T>; \
    p->add(tmp, false); \
    p->add(tmp2, nega); \
    dest = p; \
    return dest; \
  }

#define dyn2opdef(pos, posa, neg, nega,obj) \
  dyn2opdef2(pos, posa, false, obj) \
  dyn2opdef2(neg, nega, true,  obj) \
  template<typename T> \
  const dyn<T> operator pos (const T &value, const dyn<T> &r) \
  { \
    dyn<T> *tmp = new dyn<T> (new dynvalue<T> (value)); \
    obj<T> *p = new obj<T>; \
    p->add(tmp, false); \
    p->add(r, false); \
    return dyn<T> (p); \
  } \
  template<typename T> \
  const dyn<T> operator neg (const T &value, const dyn<T> &r) \
  { \
    dyn<T> *tmp = new dyn<T> (new dynvalue<T> (value)); \
    obj<T> *p = new obj<T>; \
    p->add(tmp, false); \
    p->add(r, true); \
    return dyn<T> (p); \
  }

dyn2opdef(+, +=, -, -=, dynplusminus)
dyn2opdef(*, *=, /, /=, dyntimesdiv)

#undef dyn2opdef
#undef dyn2opdef2

/* Expression base type: Modulo */
template<typename T>
class dynmod : public dyni<T>
{
private:
    const dyn<T> &L, &R;
public:
    dynmod(const dyn<T> &l, const dyn<T> &r) : L(l), R(r) { }
    virtual const T getvalue() const { return L.getvalue() % R.getvalue(); }
    virtual void debug() const { std::cout << '{'; L.debug();
                                 std::cout << '%'; R.debug();
                                 std::cout << '}'; }
    virtual const dyntype gettype() const { return DYNTYPE_MOD; }
};
template<typename T>
const dyn<T> operator % (const dyn<T> &l, const dyn<T> &r)
{
    return dyn<T> (new dynmod<T> (l, r));
}
template<typename T>
const dyn<T> &operator %= (dyn<T> &p, const dyn<T> &v)
{
    dyn<T> *q = new dyn<T>(p);
    return p = new dynmod<T> (*q, v);
}

template<typename T>
class dynfun : public dyni<T>
{
private:
    const dyn<T> &E;
    T (*const F) (T);
    const char *funame;
public:
    dynfun(const dyn<T> &exp, T (*const fun) (T), const char *f) : E(exp), F(fun), funame(f) { }
    virtual const T getvalue() const { return F(E.getvalue()); }
    virtual void debug() const { std::cout << funame << '('; E.debug(); std::cout << ')'; }
    virtual const dyntype gettype() const { return DYNTYPE_FUNC; }
};

#define dynfundef(fu) \
template<typename T> \
const dyn<T> fu(const dyn<T> &e) \
{ \
    return dyn<T> (new dynfun<T> (e, std::fu, #fu)); \
}

dynfundef(acos)
dynfundef(asin)
dynfundef(atan)
dynfundef(ceil)
dynfundef(cos)
dynfundef(cosh)
dynfundef(exp)
#if defined __USE_MISC || defined __USE_XOPEN_EXTENDED || defined __USE_ISOC99
dynfundef(expm1)
dynfundef(log1p)
dynfundef(logb)
dynfundef(cbrt)
#endif
#ifdef __USE_ISOC99
dynfundef(exp2)
dynfundef(log2)
dynfundef(round)
#endif
dynfundef(floor)
dynfundef(log)
dynfundef(log10)
dynfundef(sin)
#ifdef __USE_GNU
dynfundef(sincos)
#endif
#ifdef __USE_MISC
dynfundef(significand)
#endif
dynfundef(sinh)
dynfundef(sqrt)
dynfundef(tan)
dynfundef(tanh)

#undef dynfundef
